import pytest
import numpy as np

from CyRK.cy.cysolvertest import CySolverTester

initial_conds = np.asarray((20., 20.), dtype=np.float64)
initial_conds_complex = np.asarray((20. + 0.01j, 20. - 0.01j), dtype=np.complex128)
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-7
atol = 1.0e-8


@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_CySolverTester_resolve(rk_method, complex_valued):
    """Check that the cython class solver produced the correct result if it is re-solved multiple times"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method, auto_solve=False)

    # Solve once
    CySolverTesterInst.solve()
    assert CySolverTesterInst.success
    solution_1_t = np.copy(CySolverTesterInst.t)
    solution_1_y = np.copy(CySolverTesterInst.y)

    # Solve twice
    CySolverTesterInst.solve()
    assert CySolverTesterInst.success
    solution_2_t = np.copy(CySolverTesterInst.t)
    solution_2_y = np.copy(CySolverTesterInst.y)

    # Solve thrice
    CySolverTesterInst.solve()
    assert CySolverTesterInst.success
    solution_3_t = np.copy(CySolverTesterInst.t)
    solution_3_y = np.copy(CySolverTesterInst.y)

    # Check outputs
    # 1 and 2
    assert solution_1_t.shape == solution_2_t.shape
    assert np.all(solution_1_t == solution_2_t)
    assert solution_1_y.shape == solution_2_y.shape
    assert np.all(solution_1_y == solution_2_y)

    # 2 and 3 (if the above passed then this will implicitly check 1 and 3 too).
    assert solution_2_t.shape == solution_3_t.shape
    assert np.all(solution_2_t == solution_3_t)
    assert solution_2_y.shape == solution_3_y.shape
    assert np.all(solution_2_y == solution_3_y)


@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_CySolverTester_multi_resolve(rk_method, complex_valued):
    """Check that the cython class solver can be resolved many times without memory access issues."""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method, auto_solve=False)

    for i in range(10_000):
        # Change y0 values for each loop
        y0 = np.copy(initial_conds_to_use)
        y0[0] += 0.1
        y0[1] -= 0.1
        CySolverTesterInst.change_y0(y0, auto_reset_state=False)
        CySolverTesterInst.solve(reset=True)
        assert CySolverTesterInst.success



# distutils: language = c
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

from libc.math cimport sin, cos

from CyRK.cy.cysolver cimport CySolver


cdef class CySolverTester(CySolver):

    cdef void diffeq(self) noexcept nogil:
        
        # Unpack y
        cdef double y0, y1
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]

        self.dy_ptr[0] = (1. - 0.01 * y1) * y0
        self.dy_ptr[1] = (0.02 * y0 - 1.) * y1


cdef class CySolverAccuracyTest(CySolver):

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]

        self.dy_ptr[0] = sin(self.t_now) - y1
        self.dy_ptr[1] = cos(self.t_now) + y0


cdef class CySolverExtraTest(CySolver):

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1, extra_0, extra_1
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]

        extra_0 = (1. - 0.01 * y1)
        extra_1 = (0.02 * y0 - 1.)

        self.dy_ptr[0] = extra_0 * y0
        self.dy_ptr[1] = extra_1 * y1

        self.extra_output_ptr[0] = extra_0
        self.extra_output_ptr[1] = extra_1


cdef class CySolverLorenz(CySolver):

    cdef double a, b, c

    cdef void update_constants(self) noexcept nogil:

        self.a  = self.args_ptr[0]
        self.b  = self.args_ptr[1]
        self.c  = self.args_ptr[2]

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1, y2
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]
        y2 = self.y_ptr[2]

        self.dy_ptr[0] = self.a * (y1 - y0)
        self.dy_ptr[1] = y0 * (self.b - y2) - y1
        self.dy_ptr[2] = y0 * y1 - self.c * y2


cdef class CySolverLorenzExtra(CySolver):

    cdef double a, b, c

    cdef void update_constants(self) noexcept nogil:

        self.a  = self.args_ptr[0]
        self.b  = self.args_ptr[1]
        self.c  = self.args_ptr[2]

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1, y2, e_1, e_2, e_3
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]
        y2 = self.y_ptr[2]

        e_1 = self.a
        e_2 = (self.b - y2)
        e_3 = self.c * y2

        self.dy_ptr[0] = e_1 * (y1 - y0)
        self.dy_ptr[1] = y0 * e_2 - y1
        self.dy_ptr[2] = y0 * y1 - e_3

        self.extra_output_ptr[0] = e_1
        self.extra_output_ptr[1] = e_2
        self.extra_output_ptr[2] = e_3


cdef class CySolverLotkavolterra(CySolver):

    cdef double a, b, c, d

    cdef void update_constants(self) noexcept nogil:

        self.a = self.args_ptr[0]
        self.b = self.args_ptr[1]
        self.c = self.args_ptr[2]
        self.d = self.args_ptr[3]

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]

        self.dy_ptr[0] = self.a * y0 - self.b * y0 * y1
        self.dy_ptr[1] = -self.c * y1 + self.d * y0 * y1


cdef class CySolverPendulum(CySolver):

    cdef double coeff_1, coeff_2

    cdef void update_constants(self) noexcept nogil:

        cdef double l, m, g

        l = self.args_ptr[0]
        m = self.args_ptr[1]
        g = self.args_ptr[2]

        self.coeff_1 = (-3. * g / (2. * l))
        self.coeff_2 = (3. / (m * l**2))

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1, torque
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]

        # External torque
        torque = 0.1 * sin(self.t_now)

        self.dy_ptr[0] = y1
        self.dy_ptr[1] = self.coeff_1 * sin(y0) + self.coeff_2 * torque


cdef class CySolverStiff(CySolver):

    cdef double a

    cdef void update_constants(self) noexcept nogil:

        self.a = self.args_ptr[0]

    cdef void diffeq(self) noexcept nogil:

        # Unpack y
        cdef double y0, y1, sin_, cos_
        y0 = self.y_ptr[0]
        y1 = self.y_ptr[1]

        # External torque
        sin_ = sin(self.t_now)
        cos_ = cos(self.t_now)

        self.dy_ptr[0] = -2.0 * y0 + y1 + 2 * sin_
        self.dy_ptr[1] = (self.a - 1.0) * y0 - self.a * y1 + self.a * (cos_ - sin_)



import pytest
import numpy as np
from numba import njit

from CyRK import cyrk_ode
from CyRK.cy.cysolvertest import CySolverTester, CySolverAccuracyTest


@njit
def diffeq(t, y, dy):
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]


@njit
def diffeq_args(t, y, dy, a, b):
    dy[0] = (1. - a * y[1]) * y[0]
    dy[1] = (b * y[0] - 1.) * y[1]

initial_conds = np.asarray((20., 20.), dtype=np.float64, order='C')
initial_conds_complex = np.asarray((20. + 0.01j, 20. - 0.01j), dtype=np.complex128, order='C')
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-7
atol = 1.0e-8

rtols = np.asarray((1.0e-7, 1.0e-8), dtype=np.float64, order='C')
atols = np.asarray((1.0e-8, 1.0e-9), dtype=np.float64, order='C')

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
@pytest.mark.parametrize('use_rtol_array', (True, False))
@pytest.mark.parametrize('use_atol_array', (True, False))
def test_basic_integration_cyrk_ode(use_atol_array, use_rtol_array, rk_method, complex_valued):
    """Check that the cython function solver is able to run with its default arguments"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    if use_atol_array:
        atols_use = atols
    else:
        atols_use = None
    if use_rtol_array:
        rtols_use = rtols
    else:
        rtols_use = None

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span, initial_conds_to_use,
                 rk_method=rk_method, rtol=rtol, atol=atol, rtols=rtols_use, atols=atols_use)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
@pytest.mark.parametrize('use_rtol_array', (True, False))
@pytest.mark.parametrize('use_atol_array', (True, False))
def test_basic_integration_CySolverTester(use_atol_array, use_rtol_array, rk_method, complex_valued):
    """Check that the cython class solver is able to run with its default arguments"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    if use_atol_array:
        atols_use = atols
    else:
        atols_use = None
    if use_rtol_array:
        rtols_use = rtols
    else:
        rtols_use = None

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use,
                                        rtol=rtol, atol=atol, rtols=rtols_use, atols=atols_use,
                                        rk_method=rk_method, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str


@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_different_tols(rk_method, complex_valued):
    """Check that the cython function solver is able to run with different tolerances"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, rtol=1.0e-10, atol=1.0e-12)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_different_tols_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver is able to run with different tolerances"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method,
                                        rtol=1.0e-10, atol=1.0e-12, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_max_step(rk_method, complex_valued):
    """Check that the cython function solver is able to run with different maximum step size"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, max_step=time_span[1] / 2.)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_max_step_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver is able to run with different maximum step size"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method, max_step=time_span[1] / 2., auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_first_step(rk_method, complex_valued):
    """Check that the cython function solver is able to run with a user provided first step size"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, first_step=0.01)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_first_step_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver is able to run with a user provided first step size"""

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method, first_step=0.01, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_large_end_value(rk_method, complex_valued):
    """Check that the cython function solver is able to run using the DOP853 method. Using a larger ending time value """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span_large, initial_conds_to_use, rk_method=rk_method)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_large_end_value_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver is able to run using the DOP853 method. Using a larger ending time value """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span_large, initial_conds_to_use, rk_method=rk_method, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_teval(rk_method, complex_valued):
    """Check that the cython function solver is able to run using a user provided t_eval """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    t_eval = np.linspace(time_span[0], time_span[1], 10)

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span, initial_conds_to_use, rk_method=rk_method, t_eval=t_eval)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == t_eval.size
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_teval_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver is able to run using a user provided t_eval """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    t_eval = np.linspace(time_span[0], time_span[1], 10)

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method, t_eval=t_eval, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_args(rk_method, complex_valued):
    """Check that the cython function solver is able to run with user provided additional diffeq arguments """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    time_domain, y_results, success, message = \
        cyrk_ode(diffeq_args, time_span, initial_conds_to_use, rk_method=rk_method, args=(0.01, 0.02))

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_args_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver is able to run with user provided additional diffeq arguments """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    CySolverTesterInst = CySolverTester(time_span, initial_conds_to_use, rk_method=rk_method, args=(0.01, 0.02), auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_accuracy(rk_method):
    """Check that the cython function solver is able to reproduce a known functions integral with reasonable accuracy """

    # TODO: This is only checking one equation. Add other types of diffeqs to provide better coverage.

    # Differential Equation
    @njit
    def diffeq_accuracy(t, y, dy):
        dy[0] = np.sin(t) - y[1]  # dydt = sin(t) - x(t)
        dy[1] = np.cos(t) + y[0]  # dxdt = cos(t) + y(t)

    @njit
    def correct_answer(t, c1_, c2_):
        y = np.empty((2, t.size), dtype=np.float64)
        y[0] = -c1_ * np.sin(t) + c2_ * np.cos(t) - (np.cos(t) / 2)  # -c1 * sin(t) + c2 * cos(t) - cos(t) / 2
        # At t=0; y = c2 - 1/2
        y[1] = c2_ * np.sin(t) + c1_ * np.cos(t) + (np.sin(t) / 2)   # c2 * sin(t) + c1 * cos(t) + sin(t) / 2
        # At t=0; x = c1
        return y

    # Initial Conditions
    # y=0 --> c2 = + 1/2
    c2 = 0.5
    # x=1 --> c1 = + 1
    c1 = 1.0
    y0 = np.asarray((0., 1.), dtype=np.float64)
    time_span_ = (0., 10.)

    # CyRK.cyrk_ode
    time_domain, y_results, success, message = \
        cyrk_ode(diffeq_accuracy, time_span_, y0, rk_method=rk_method, rtol=1.0e-8, atol=1.0e-9)
    real_answer = correct_answer(time_domain, c1, c2)

    if rk_method == 0:
        assert np.allclose(y_results, real_answer, rtol=1.0e-3, atol=1.0e-6)
    elif rk_method == 1:
        assert np.allclose(y_results, real_answer, rtol=1.0e-4, atol=1.0e-7)
    else:
        assert np.allclose(y_results, real_answer, rtol=1.0e-5, atol=1.0e-8)

    # Check the accuracy of the results
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # ax.plot(time_domain, y_results[0], 'r', label='CyRK')
    # ax.plot(time_domain, y_results[1], 'r:')
    # ax.plot(time_domain, real_answer[0], 'b', label='Analytic')
    # ax.plot(time_domain, real_answer[1], 'b:')
    # plt.show()

@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_accuracy_CySolverTester(rk_method):
    """Check that the cython class solver is able to reproduce a known functions integral with reasonable accuracy """

    # TODO: This is only checking one equation. Add other types of diffeqs to provide better coverage.

    @njit
    def correct_answer(t, c1_, c2_):
        y = np.empty((2, t.size), dtype=np.float64)
        y[0] = -c1_ * np.sin(t) + c2_ * np.cos(t) - (np.cos(t) / 2)  # -c1 * sin(t) + c2 * cos(t) - cos(t) / 2
        # At t=0; y = c2 - 1/2
        y[1] = c2_ * np.sin(t) + c1_ * np.cos(t) + (np.sin(t) / 2)   # c2 * sin(t) + c1 * cos(t) + sin(t) / 2
        # At t=0; x = c1
        return y

    # Initial Conditions
    # y=0 --> c2 = + 1/2
    c2 = 0.5
    # x=1 --> c1 = + 1
    c1 = 1.0
    y0 = np.asarray((0., 1.), dtype=np.float64)
    time_span_ = (0., 10.)

    # CyRK.CySolver
    CySolverAccuracyTestInst = CySolverAccuracyTest(time_span_, y0, rk_method=rk_method, rtol=1.0e-8, atol=1.0e-9, auto_solve=True)
    real_answer = correct_answer(CySolverAccuracyTestInst.t, c1, c2)

    if rk_method == 0:
        assert np.allclose(CySolverAccuracyTestInst.y, real_answer, rtol=1.0e-3, atol=1.0e-4)
    elif rk_method == 1:
        assert np.allclose(CySolverAccuracyTestInst.y, real_answer, rtol=1.0e-4, atol=1.0e-5)
    else:
        assert np.allclose(CySolverAccuracyTestInst.y, real_answer, rtol=1.0e-5, atol=1.0e-6)

    # Check the accuracy of the results
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # ax.plot(CySolverAccuracyTestInst.t, CySolverAccuracyTestInst.y[0], 'r', label='CyRK')
    # ax.plot(CySolverAccuracyTestInst.t, CySolverAccuracyTestInst.y[1], 'r:')
    # ax.plot(CySolverAccuracyTestInst.t, real_answer[0], 'b', label='Analytic')
    # ax.plot(CySolverAccuracyTestInst.t, real_answer[1], 'b:')
    # plt.show()

import platform
if platform.system().lower() == 'darwin':
    # TODO: For some reason the accuracy test is failing on macos after the switch from GCC. 
    # Since this is a deprecated function there is not a strong incentive to investigate since the new backend passes.
    # But something to keep in mind if we ever revive it.
    pytest.mark.skip("For some reason the accuracy test is failing on macos after the switch from GCC. Since this is a deprecated function there is not a strong incentive to investigate since the new backend passes. But something to keep in mind if we ever revive it.")(test_accuracy_CySolverTester)

@pytest.mark.parametrize('complex_valued', (True, False))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_max_num_steps(rk_method, complex_valued):
    """Check that the cython function cyrk_ode can use max_num_steps argument """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    # First test a number of max steps which is fine.
    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span_large, initial_conds_to_use, rk_method=rk_method, max_num_steps=1000000)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    if complex_valued:
        assert y_results.dtype == np.complex128
    else:
        assert y_results.dtype == np.float64
    assert time_domain.size > 1
    assert time_domain.size == y_results[0].size
    assert len(y_results.shape) == 2
    assert y_results[0].size == y_results[1].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

    # Now test an insufficient number of steps
    time_domain, y_results, success, message = \
        cyrk_ode(diffeq, time_span_large, initial_conds_to_use, rk_method=rk_method, max_num_steps=4)

    assert not success
    assert message == "Maximum number of steps (set by user) exceeded during integration."

@pytest.mark.parametrize('complex_valued', (False,))
@pytest.mark.parametrize('rk_method', (0, 1, 2))
def test_max_num_steps_CySolverTester(rk_method, complex_valued):
    """Check that the cython class solver correctly uses the max_num_steps argument. """

    if complex_valued:
        initial_conds_to_use = initial_conds_complex
    else:
        initial_conds_to_use = initial_conds

    # First test a number of max steps which is fine.
    CySolverTesterInst = CySolverTester(time_span_large, initial_conds_to_use, rk_method=rk_method, auto_solve=True, max_num_steps=1000000)

    # Check that the ndarrays make sense
    assert type(CySolverTesterInst.t) == np.ndarray
    assert CySolverTesterInst.t.dtype == np.float64
    if complex_valued:
        assert CySolverTesterInst.y.dtype == np.complex128
    else:
        assert CySolverTesterInst.y.dtype == np.float64
    assert CySolverTesterInst.t.size > 1
    assert CySolverTesterInst.t.size == CySolverTesterInst.y[0].size
    assert len(CySolverTesterInst.y.shape) == 2
    assert CySolverTesterInst.y[0].size == CySolverTesterInst.y[1].size

    # Check that the other output makes sense
    assert type(CySolverTesterInst.success) == bool
    assert CySolverTesterInst.success
    assert type(CySolverTesterInst.message) == str

    # Now test an insufficient number of steps
    CySolverTesterInst = CySolverTester(
        time_span_large, initial_conds_to_use, rk_method=rk_method, auto_solve=True, max_num_steps=4)

    assert not CySolverTesterInst.success
    assert CySolverTesterInst.status == -2

def test_bad_tols_cyrk():

    # Too many rtols and atols
    bad_rtols = np.asarray((1.0e-6, 1.0e-7, 1.0e-8), dtype=np.float64, order='C')
    bad_atols = np.asarray((1.0e-7, 1.0e-8, 1.0e-9), dtype=np.float64, order='C')

    with pytest.raises(AttributeError):
        time_domain, y_results, success, message = \
            cyrk_ode(diffeq, time_span_large, initial_conds, rk_method=1, rtols=bad_rtols)

    with pytest.raises(AttributeError):
        time_domain, y_results, success, message = \
            cyrk_ode(diffeq, time_span_large, initial_conds, rk_method=1, atols=bad_atols)


def test_bad_tols_CySolver():

    # Too many rtols and atols
    bad_rtols = np.asarray((1.0e-6, 1.0e-7, 1.0e-8), dtype=np.float64, order='C')
    bad_atols = np.asarray((1.0e-7, 1.0e-8, 1.0e-9), dtype=np.float64, order='C')

    with pytest.raises(AttributeError):
        CySolverTesterInst = CySolverTester(time_span_large, initial_conds,
                                            rk_method=1, rtols=bad_rtols, auto_solve=True)

    with pytest.raises(AttributeError):
        CySolverTesterInst = CySolverTester(time_span_large, initial_conds,
                                            rk_method=1, atols=bad_atols, auto_solve=True)


if __name__ == '__main__':

    test_basic_integration_cyrk_ode(False, False, 2, False)



# distutils: language = c
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

import cython

import numpy as np
cimport numpy as np

from libc.math cimport sqrt, fabs, nextafter, NAN, floor

from CyRK.utils.utils cimport allocate_mem, reallocate_mem, free_mem
from CyRK.rk.rk cimport find_rk_properties
from CyRK.cy.common cimport double_numeric, interpolate, SAFETY, MIN_FACTOR, MAX_FACTOR, MAX_STEP, INF, \
    EPS_100, find_expected_size, find_max_num_steps

import warnings

cdef double cabs(
        double complex value
        ) noexcept nogil:
    """ Absolute value function for complex-valued inputs.

    Parameters
    ----------
    value : float (double complex)
        Complex-valued number.

    Returns
    -------
    value_abs : float (double)
        Absolute value of `value`.
    """

    cdef double v_real
    cdef double v_imag
    v_real = value.real
    v_imag = value.imag

    return sqrt(v_real * v_real + v_imag * v_imag)


cdef double dabs(
        double_numeric value
        ) noexcept nogil:
    """ Absolute value function for either float or complex-valued inputs.

    Checks the type of value and either utilizes `cabs` (for double complex) or `fabs` (for floats).

    Parameters
    ----------
    value : float (double_numeric)
        Float or complex-valued number.

    Returns
    -------
    value_abs : float (double)
        Absolute value of `value`.
    """

    # Check the type of value
    if double_numeric is cython.doublecomplex:
        return cabs(value)
    else:
        return fabs(value)


def cyrk_ode(
        diffeq,
        (double, double) t_span,
        const double_numeric[:] y0,
        tuple args = None,
        double rtol = 1.e-3,
        double atol = 1.e-6,
        double[::1] rtols = None,
        double[::1] atols = None,
        double max_step = MAX_STEP,
        double first_step = 0.,
        unsigned char rk_method = 1,
        double[:] t_eval = None,
        bint capture_extra = False,
        size_t num_extra = 0,
        bint interpolate_extra = False,
        size_t expected_size = 0,
        size_t max_num_steps = 0,
        size_t max_ram_MB = 2000,
        bint raise_warnings = True
        ):
    """
    cyrk_ode: A Runge-Kutta Solver Implemented in Cython.

    Parameters
    ----------
    diffeq : callable
        A python or njit-ed numba differential equation.
        Format should follow:
        ```
        def diffeq(t, y, dy, arg_1, arg_2, ...):
            dy[0] = y[0] * t
            ....
        ```
    t_span : (double, double)
        Values of independent variable at beginning and end of integration.
    y0 : double[::1]
        Initial values for the dependent y variables at `t_span[0]`.
    args : tuple or None, default=None
        Additional arguments used by the differential equation.
        None (default) will tell the solver to not use additional arguments.
    rk_method : int, default=1
        Runge-Kutta method that will be used. Currently implemented models:
            0: ‘RK23’: Explicit Runge-Kutta method of order 3(2).
            1: ‘RK45’ (default): Explicit Runge-Kutta method of order 5(4).
            2: ‘DOP853’: Explicit Runge-Kutta method of order 8.
    rtol : double, default=1.0e-3
        Relative tolerance using in local error calculation.
    atol : double, default=1.0e-6
        Absolute tolerance using in local error calculation.
    rtols : double[::1], default=None
        np.ndarray of relative tolerances, one for each dependent y variable.
        None (default) will use the same tolerance (set by `rtol`) for each y variable.
    atols : double[::1], default=None
        np.ndarray of absolute tolerances, one for each dependent y variable.
        None (default) will use the same tolerance (set by `atol`) for each y variable.
    max_step : double, default=+Inf
        Maximum allowed step size.
    first_step : double, default=0
        First step's size (after `t_span[0]`).
        If set to 0 (the default) then the solver will attempt to guess a suitable initial step size.
    max_num_steps : size_t, default=0
        Maximum number of step sizes allowed before solver will auto fail.
        If set to 0 (the default) then the maximum number of steps will be equal to max integer size
        allowed on system architecture.
    t_eval : double[::1], default=None
        If not set to None, then a final interpolation will be performed on the solution to fit it to this array.
    capture_extra : bool = False
        If True, then additional output from the differential equation will be collected (but not used to determine
         integration error).
         Example:
            ```
            def diffeq(t, y, dy):
                a = ... some function of y and t.
                dy[0] = a**2 * sin(t) - y[1]
                dy[1] = a**3 * cos(t) + y[0]

                # Storing extra output in dy even though it is not part of the diffeq.
                dy[2] = a
            ```
    num_extra : int = 0
        The number of extra outputs the integrator should expect. With the previous example there is 1 extra output.
    interpolate_extra : bint, default=False
        Flag if interpolation should be run on extra parameters.
        If set to False when `run_interpolation=True`, then interpolation will be run on solution's y, t. These will
        then be used to recalculate extra parameters rather than an interpolation on the extra parameters captured
        during integration.
    expected_size : size_t, default=0
        Anticipated size of integration range, i.e., how many steps will be required.
        Used to build temporary storage arrays for the solution results.
        If set to 0 (the default), then the solver will attempt to guess on a suitable expected size based on the
        relative tolerances and size of the integration domain.

    Returns
    -------
    time_domain : np.ndarray
        The final time domain. This is equal to t_eval if it was provided.
    y_results : np.ndarray
        The solution of the differential equation provided for each time_result.
    success : bool
        Final integration success flag.
    message : str
        Any integration messages, useful if success=False.
    """

    if raise_warnings:
        warnings.warn(
            "`cyrk_ode` method is now deprecated it will be removed in the next major update of CyRK. "
            "Please see the documentation on the new `pysolve_ivp` function which acts as its replacement.",
            DeprecationWarning
            )

    # Setup loop variables
    cdef size_t s, i, j

    # Setup integration variables
    cdef char status
    cdef str message

    # Determine information about the differential equation based on its initial conditions
    cdef size_t y_size
    cdef double y_size_dbl, y_size_sqrt
    cdef bint y_is_complex
    y_size       = y0.size
    y_is_complex = False
    y_size_dbl   = <double>y_size
    y_size_sqrt  = sqrt(y_size_dbl)

    # Check the type of the values in y0
    if double_numeric is cython.double:
        DTYPE = np.float64
    elif double_numeric is cython.doublecomplex:
        DTYPE = np.complex128
        y_is_complex = True
    else:
        # Cyrk only supports float64 and complex128.
        status = -8
        message = "Attribute error."
        raise Exception('Unexpected type found for initial conditions (y0).')

    # Build time domain
    cdef double t_start, t_end, t_delta, t_delta_check, t_delta_abs, direction_inf, t_old, t_now, time_
    cdef bint direction_flag
    t_start       = t_span[0]
    t_end         = t_span[1]
    t_delta       = t_end - t_start
    t_delta_abs   = fabs(t_delta)
    t_delta_check = t_delta_abs

    if t_delta >= 0.:
        # Integration is moving forward in time.
        direction_flag = True
        direction_inf = INF
    else:
        # Integration is moving backwards in time.
        direction_flag = False
        direction_inf = -INF

    # Pull out information on args
    cdef bint use_args
    if args is None:
        use_args = False
    else:
        use_args = True

    # Setup temporary variables to store intermediate values
    cdef double temp_double
    cdef double_numeric temp_double_numeric

    # Determine integration tolerances
    cdef double* tol_ptrs = NULL
    cdef double* rtols_ptr = NULL
    cdef double* atols_ptr = NULL
    tol_ptrs  = <double *> allocate_mem(2 * y_size * sizeof(double), 'tol_ptrs (start-up)')
    rtols_ptr = &tol_ptrs[0]
    atols_ptr = &tol_ptrs[y_size]
    
    cdef double rtol_min
    rtol_min = INF
    if rtols is not None:
        # User provided an arrayed version of rtol.
        if len(rtols) != y_size:
            raise AttributeError('rtol array must be the same size as y0.')
        for i in range(y_size):
            temp_double = rtols[i]
            # Check that the tolerances are not too small.
            if temp_double < EPS_100:
                temp_double = EPS_100
            rtol_min = min(rtol_min, temp_double)
            rtols_ptr[i] = temp_double
    else:
        # No array provided. Use the same rtol for all ys.
        # Check that the tolerances are not too small.
        if rtol < EPS_100:
            rtol = EPS_100
        rtol_min = rtol
        for i in range(y_size):
            rtols_ptr[i] = rtol

    if atols is not None:
        # User provided an arrayed version of atol.
        if len(atols) != y_size:
            raise AttributeError('atol array must be the same size as y0.')
        for i in range(y_size):
            atols_ptr[i] = atols[i]
    else:
        # No array provided. Use the same atol for all ys.
        for i in range(y_size):
            atols_ptr[i] = atol
    
    # Determine max number of steps
    cdef size_t max_num_steps_touse
    cdef bint user_provided_max_num_steps
    find_max_num_steps(
        y_size,
        num_extra,
        max_num_steps,
        max_ram_MB,
        capture_extra,
        y_is_complex,
        &user_provided_max_num_steps,
        &max_num_steps_touse)

    # Expected size of output arrays.
    cdef size_t expected_size_to_use, num_concats, current_size
    if expected_size == 0:
        # cyrk_ode will attempt to guess on a best size for the arrays.
        expected_size_to_use = find_expected_size(
                y_size,
                num_extra,
                t_delta_abs,
                rtol_min,
                capture_extra,
                y_is_complex)
    else:
        expected_size_to_use = expected_size
    # Set the current size to the expected size.
    # `expected_size` should never change but current might grow if expected size is not large enough.
    current_size = expected_size_to_use
    num_concats  = 1

    # Initialize live variable arrays
    cdef double_numeric* y_storage_ptrs = NULL
    cdef double_numeric* y_old_ptr = NULL
    cdef double_numeric* dy_ptr = NULL
    cdef double_numeric* dy_old_ptr = NULL

    y_storage_ptrs = <double_numeric *> allocate_mem(3 * y_size * sizeof(double_numeric), 'y_storage_ptrs (start-up)')

    y_old_ptr  = &y_storage_ptrs[0]
    dy_ptr     = &y_storage_ptrs[1 * y_size]
    dy_old_ptr = &y_storage_ptrs[2 * y_size]

    # Build memoryviews based on y_view and dy_ptr that can be passed to the diffeq.
    # This is process is different than CySolver which strictly uses c pointers.
    # These memoryviews allow for user-provided diffeqs that are not cython/compiled.
    y_array = np.empty(y_size, dtype=DTYPE, order='C')
    cdef double_numeric[::1] y_view = y_array

    # Store y0 into the y arrays
    for i in range(y_size):
        temp_double_numeric = y0[i]
        y_view[i]    = temp_double_numeric
        y_old_ptr[i] = temp_double_numeric

    # Determine extra outputs
    # To avoid memory access violations we need to set the extra output arrays no matter if they are used.
    # If not used, just set them to size zero.
    if capture_extra:
        if num_extra <= 0:
            status = -8
            raise AttributeError('Capture extra set to True, but number of extra set to 0 (or negative).')
    else:
        # Even though we are not capturing extra, we still want num_extra to be equal to 1 so that nan arrays
        # are properly initialized
        num_extra = 1

    cdef double_numeric* extra_output_init_ptr = NULL
    cdef double_numeric* extra_output_ptr = NULL
    extra_output_init_ptr = <double_numeric *> allocate_mem(
        num_extra * sizeof(double_numeric),
        'extra_output_init_ptr (start-up)')
    extra_output_ptr = <double_numeric *> allocate_mem(
        num_extra * sizeof(double_numeric),
        'extra_output_ptr (start-up)')

    for i in range(num_extra):
        extra_output_init_ptr[i] = NAN
        extra_output_ptr[i]      = NAN

    # If extra output is true then the output of the diffeq will be larger than the size of y0.
    # Determine that extra size by calling the diffeq and checking its size.
    cdef size_t extra_start, total_size
    extra_start = y_size
    if capture_extra:
        total_size = y_size + num_extra
    else:
        total_size = y_size

    # Build pointer to store results of diffeq
    diffeq_out_array = np.empty(total_size, dtype=DTYPE, order='C')
    cdef double_numeric[::1] diffeq_out_view = diffeq_out_array

    # Determine interpolation information
    cdef bint run_interpolation
    cdef size_t len_t_eval
    if t_eval is None:
        run_interpolation = False
        interpolate_extra = False
        # Even though we are not using t_eval, set its size equal to one so that nan arrays can be built
        len_t_eval = 1
    else:
        run_interpolation = True
        interpolate_extra = interpolate_extra
        len_t_eval = len(t_eval)

    cdef double* t_eval_ptr = NULL
    t_eval_ptr = <double *> allocate_mem(len_t_eval * sizeof(double), 't_eval_ptr (start-up)')
    for i in range(len_t_eval):
        if run_interpolation:
            t_eval_ptr[i] = t_eval[i]
        else:
            t_eval_ptr[i] = NAN

    # Make initial call to diffeq to get initial dydt and any extra outputs (if requested) at t0.
    if use_args:
        diffeq(t_start, y_array, diffeq_out_array, *args)
    else:
        diffeq(t_start, y_array, diffeq_out_array)

    # Setup initial conditions
    t_old = t_start
    t_now = t_start
    for i in range(y_size):
        temp_double_numeric = diffeq_out_view[i]
        dy_ptr[i]     = temp_double_numeric
        dy_old_ptr[i] = temp_double_numeric

    # Capture the extra output for the initial condition.
    if capture_extra:
        for i in range(num_extra):
            # Pull from extra output
            extra_output_init_ptr[i] = diffeq_out_view[extra_start + i]

    # Determine RK scheme and initialize RK memory views
    cdef double* A_ptr = NULL
    cdef double* B_ptr = NULL
    cdef double* C_ptr = NULL
    cdef double* E_ptr = NULL
    cdef double* E3_ptr = NULL
    cdef double* E5_ptr = NULL
    cdef size_t rk_order, error_order, rk_n_stages, len_Arows, len_Acols, len_C, rk_n_stages_plus1
    cdef double error_expo, error_pow

    find_rk_properties(
        rk_method,
        &rk_order,
        &error_order,
        &rk_n_stages,
        &len_Arows,
        &len_Acols,
        &A_ptr,
        &B_ptr,
        &C_ptr,
        &E_ptr,
        &E3_ptr,
        &E5_ptr
        )

    if rk_order == 0:
        raise AttributeError('Unknown or not-yet-implemented RK method requested.')
    
    len_C             = rk_n_stages
    rk_n_stages_plus1 = rk_n_stages + 1
    error_expo        = 1. / (<double>error_order + 1.)
    
    # Initialize other RK-related Arrays
    cdef double_numeric* K_ptr = NULL
    K_ptr = <double_numeric *> allocate_mem(rk_n_stages_plus1 * y_size * sizeof(double_numeric), 'K_ptr (start-up)')
    # It is important K be initialized with 0s
    for i in range(rk_n_stages_plus1):
        for j in range(y_size):
            K_ptr[i * y_size + j] = 0.
    
    cdef double error_norm5, error_norm3, error_norm, error_norm_abs, error_norm3_abs, error_norm5_abs, error_denom

    # Other RK Optimizations
    cdef double A_at_sj, A_at_10, B_at_j
    cdef double_numeric K_
    A_at_10 = A_ptr[1 * len_Acols + 0]

    # Setup storage arrays
    # These arrays are built to fit a number of points equal to current_size
    # If the integration needs more than that then a new array will be concatenated (with performance costs) to these.
    cdef double* time_domain_array_ptr = NULL
    cdef double_numeric* y_results_array_ptr = NULL
    cdef double_numeric* extra_array_ptr = NULL

    time_domain_array_ptr = <double *> allocate_mem(
        current_size * sizeof(double),
        'time_domain_array_ptr (start-up)')
    y_results_array_ptr = <double_numeric *> allocate_mem(
        y_size * current_size * sizeof(double_numeric),
        'y_results_array_ptr (start-up)')
    if capture_extra:
        extra_array_ptr = <double_numeric *> allocate_mem(
            num_extra * current_size * sizeof(double_numeric),
            'extra_array_ptr (start-up)')

    # Load initial conditions into storage arrays
    time_domain_array_ptr[0] = t_start
    for i in range(y_size):
        y_results_array_ptr[i] = y0[i]
    if capture_extra:
        for i in range(num_extra):
            extra_array_ptr[i] = extra_output_init_ptr[i]

    # Solution pointers
    cdef double* solution_t_ptr = NULL
    cdef double_numeric* solution_y_ptr = NULL
    cdef double_numeric* solution_extra_ptr = NULL

    # Determine size of first step.
    cdef double d0, d1, d2, d0_abs, d1_abs, d2_abs, h0, h1, scale
    cdef double step, step_size, min_step, step_factor

    # Integration flags and variables
    cdef bint success, step_accepted, step_rejected, step_error
    cdef size_t len_t

    # Integration completion variables
    cdef size_t len_t_touse
    cdef double* interpolated_solution_t_ptr = NULL
    cdef double_numeric* interpolated_solution_y_ptr = NULL
    cdef double_numeric* interpolated_solution_extra_ptr = NULL
    cdef np.ndarray[np.float64_t, ndim=1, mode='c'] solution_t
    cdef np.ndarray[double_numeric, ndim=2, mode='c'] solution_y
    cdef double[::1] solution_t_view
    cdef double_numeric[:, ::1] solution_y_view

    try:
        if first_step == 0.:
            # Select an initial step size based on the differential equation.
            # .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
            #        Equations I: Nonstiff Problems", Sec. II.4.
            if y_size == 0:
                step_size = INF
            else:
                # Find the norm for d0 and d1
                d0 = 0.
                d1 = 0.
                for i in range(y_size):

                    temp_double = dabs(y_old_ptr[i])
                    scale = atols_ptr[i] + dabs(temp_double) * rtols_ptr[i]
                    d0_abs = dabs(temp_double) / scale
                    d1_abs = dabs(dy_old_ptr[i]) / scale
                    d0 += (d0_abs * d0_abs)
                    d1 += (d1_abs * d1_abs)

                d0 = sqrt(d0) / y_size_sqrt
                d1 = sqrt(d1) / y_size_sqrt

                if d0 < 1.e-5 or d1 < 1.e-5:
                    h0 = 1.e-6
                else:
                    h0 = 0.01 * d0 / d1

                if direction_flag:
                    h0_direction = h0
                else:
                    h0_direction = -h0
                t_now = t_old + h0_direction
                for i in range(y_size):
                    y_view[i] = y_old_ptr[i] + h0_direction * dy_old_ptr[i]

                if use_args:
                    diffeq(t_now, y_array, diffeq_out_array, *args)
                else:
                    diffeq(t_now, y_array, diffeq_out_array)

                # Find the norm for d2
                d2 = 0.
                for i in range(y_size):
                    temp_double_numeric = diffeq_out_view[i]
                    dy_ptr[i] = temp_double_numeric
                    scale = atols_ptr[i] + dabs(y_old_ptr[i]) * rtols_ptr[i]
                    d2_abs = dabs( (temp_double_numeric - dy_old_ptr[i]) ) / scale
                    d2 += (d2_abs * d2_abs)

                d2 = sqrt(d2) / (h0 * y_size_sqrt)

                if d1 <= 1.e-15 and d2 <= 1.e-15:
                    h1 = max(1.e-6, h0 * 1.e-3)
                else:
                    h1 = (0.01 / max(d1, d2))**error_expo

                step_size = max(10. * fabs(nextafter(t_old, direction_inf) - t_old), min(100. * h0, h1))
        else:
            if first_step <= 0.:
                status = -8
                message = "Attribute error."
                raise AttributeError('Error in user-provided step size: Step size must be a positive number.')
            elif first_step > t_delta_abs:
                status = -8
                message = "Attribute error."
                raise AttributeError('Error in user-provided step size: Step size can not exceed bounds.')
            step_size = first_step

        # # Main integration loop
        # Set integration flags
        success       = False
        step_accepted = False
        step_rejected = False
        step_error    = False
        status        = 0
        message       = "Integration is/was ongoing (perhaps it was interrupted?)."

        # Track number of steps.
        # Initial conditions were provided so the number of steps is already 1
        len_t = 1

        if y_size == 0:
            status = -6
            message = "Integration never started: y-size is zero."

        while status == 0:
            if t_now == t_end:
                t_old = t_end
                status = 1
                break

            if len_t > max_num_steps_touse:
                if user_provided_max_num_steps:
                    status = -2
                    message = "Maximum number of steps (set by user) exceeded during integration."
                else:
                    status = -3
                    message = "Maximum number of steps (set by ram usage limit) exceeded during integration."
                break

            # Run RK integration step
            # Determine step size based on previous loop
            # Find minimum step size based on the value of t (less floating point numbers between numbers when t is large)
            min_step = 10. * fabs(nextafter(t_old, direction_inf) - t_old)
            # Look for over/undershoots in previous step size
            if step_size > max_step:
                step_size = max_step
            elif step_size < min_step:
                step_size = min_step

            # Determine new step size
            step_accepted = False
            step_rejected = False
            step_error    = False

            # # Step Loop
            while not step_accepted:

                if step_size < min_step:
                    step_error = True
                    status     = -1
                    break

                # Move time forward for this particular step size
                if direction_flag:
                    step = step_size
                    t_now = t_old + step
                    t_delta_check = t_now - t_end
                else:
                    step = -step_size
                    t_now = t_old + step
                    t_delta_check = t_end - t_now

                # Check that we are not at the end of integration with that move
                if t_delta_check > 0.:
                    t_now = t_end

                    # Correct the step if we were at the end of integration
                    step = t_now - t_old
                    if direction_flag:
                        step_size = step
                    else:
                        step_size = -step

                # Calculate derivative using RK method
                # Dot Product (K, a) * step
                for s in range(1, len_C):
                    time_ = t_old + C_ptr[s] * step

                    # Dot Product (K, a) * step
                    if s == 1:
                        for i in range(y_size):
                            # Set the first column of K
                            temp_double_numeric = dy_old_ptr[i]
                            K_ptr[i] = temp_double_numeric

                            # Calculate y_new for s==1
                            y_view[i] = y_old_ptr[i] + (temp_double_numeric * A_at_10 * step)
                    else:
                        for j in range(s):
                            A_at_sj = A_ptr[s * len_Acols + j] * step
                            for i in range(y_size):
                                if j == 0:
                                    # Initialize
                                    y_view[i] = y_old_ptr[i]

                                y_view[i] = y_view[i] + K_ptr[j * y_size + i] * A_at_sj

                    if use_args:
                        diffeq(time_, y_array, diffeq_out_array, *args)
                    else:
                        diffeq(time_, y_array, diffeq_out_array)

                    for i in range(y_size):
                        K_ptr[s * y_size + i] = diffeq_out_view[i]

                # Dot Product (K, B) * step
                for j in range(rk_n_stages):
                    B_at_j = B_ptr[j] * step
                    # We do not use rk_n_stages_plus1 here because we are chopping off the last row of K to match
                    #  the shape of B.
                    for i in range(y_size):
                        if j == 0:
                            # Initialize
                            y_view[i] = y_old_ptr[i]

                        y_view[i] = y_view[i] + K_ptr[j * y_size + i] * B_at_j

                # Find final dydt for this timestep
                if use_args:
                    diffeq(t_now, y_array, diffeq_out_array, *args)
                else:
                    diffeq(t_now, y_array, diffeq_out_array)

                # Store extra
                if capture_extra:
                    for i in range(num_extra):
                        extra_output_ptr[i] = diffeq_out_view[extra_start + i]

                if rk_method == 2:
                    # Calculate Error for DOP853
                    # Find norms for each error
                    error_norm5 = 0.
                    error_norm3 = 0.
                    # Dot Product (K, E5) / scale and Dot Product (K, E3) * step / scale
                    for i in range(y_size):
                        # Find scale of y for error calculations
                        scale = atols_ptr[i] + max(dabs(y_old_ptr[i]), dabs(y_view[i])) * rtols_ptr[i]

                        # Set diffeq results
                        temp_double_numeric = diffeq_out_view[i]
                        dy_ptr[i] = temp_double_numeric

                        # Set last array of K equal to dydt
                        K_ptr[rk_n_stages * y_size + i] = temp_double_numeric
                        # Initialize
                        error_dot_1 = 0.
                        error_dot_2 = 0.
                        for j in range(rk_n_stages_plus1):

                            K_ = K_ptr[j * y_size + i]
                            error_dot_1 += K_ * E3_ptr[j]
                            error_dot_2 += K_ * E5_ptr[j]

                        error_norm3_abs = dabs(error_dot_1) / scale
                        error_norm5_abs = dabs(error_dot_2) / scale

                        error_norm3 += (error_norm3_abs * error_norm3_abs)
                        error_norm5 += (error_norm5_abs * error_norm5_abs)

                    # Check if errors are zero
                    if (error_norm5 == 0.) and (error_norm3 == 0.):
                        error_norm = 0.
                    else:
                        error_denom = error_norm5 + 0.01 * error_norm3
                        error_norm  = step_size * error_norm5 / sqrt(error_denom * y_size_dbl)

                else:
                    # Calculate Error for RK23 and RK45
                    # Dot Product (K, E) * step / scale
                    error_norm = 0.
                    for i in range(y_size):
                        # Find scale of y for error calculations
                        scale = atols_ptr[i] + max(dabs(y_old_ptr[i]), dabs(y_view[i])) * rtols_ptr[i]

                        # Set diffeq results
                        temp_double_numeric = diffeq_out_view[i]
                        dy_ptr[i] = temp_double_numeric

                        # Set last array of K equal to dydt
                        K_ptr[rk_n_stages * y_size + i] = temp_double_numeric
                        # Initialize
                        error_dot_1 = 0.
                        for j in range(rk_n_stages_plus1):

                            error_dot_1 += K_ptr[j * y_size + i] * E_ptr[j]

                        error_norm_abs = dabs(error_dot_1) * (step / scale)
                        error_norm    += (error_norm_abs * error_norm_abs)
                    error_norm = sqrt(error_norm) / y_size_sqrt

                if error_norm < 1.:
                    # The error is low! Let's update this step for the next time loop
                    if error_norm == 0.:
                        step_factor = MAX_FACTOR
                    else:
                        error_pow = error_norm**-error_expo
                        step_factor = min(MAX_FACTOR, SAFETY * error_pow)

                    if step_rejected:
                        # There were problems with this step size on the previous step loop. Make sure factor does
                        #    not exasperate them.
                        step_factor = min(step_factor, 1.)

                    step_size = step_size * step_factor
                    step_accepted = True
                else:
                    error_pow = error_norm**-error_expo
                    step_size = step_size * max(MIN_FACTOR, SAFETY * error_pow)
                    step_rejected = True

            if step_error:
                # Issue with step convergence
                status = -1
                message = "Error in step size calculation:\n\tRequired step size is less than spacing between numbers."
                break
            elif not step_accepted:
                # Issue with step convergence
                status = -7
                message = "Error in step size calculation:\n\tError in step size acceptance."
                break

            # End of step loop. Update the _now variables
            t_old = t_now
            for i in range(y_size):
                y_old_ptr[i] = y_view[i]
                dy_old_ptr[i] = dy_ptr[i]

            # Store data
            if len_t >= current_size:
                # There is more data then we have room in our arrays.
                # Build new arrays with more space.
                # OPT: Note this is an expensive operation.
                num_concats += 1

                # Grow the array by 50% its current value
                current_size = <size_t> floor(<double>current_size * (1.5))

                time_domain_array_ptr = <double *> reallocate_mem(
                    time_domain_array_ptr,
                    current_size * sizeof(double),
                    'time_domain_array_ptr (growth stage)')

                y_results_array_ptr = <double_numeric *> reallocate_mem(
                    y_results_array_ptr,
                    y_size * current_size * sizeof(double_numeric),
                    'y_results_array_ptr (growth stage)')

                if capture_extra:
                    extra_array_ptr = <double_numeric *> reallocate_mem(
                        extra_array_ptr,
                        num_extra * current_size * sizeof(double_numeric),
                        'extra_array_ptr (growth stage)')

            # Add this step's results to our storage arrays.
            time_domain_array_ptr[len_t] = t_now
            for i in range(y_size):
                y_results_array_ptr[len_t * y_size + i] = y_view[i]

            if capture_extra:
                for i in range(num_extra):
                    extra_array_ptr[len_t * num_extra + i] = extra_output_ptr[i]

            # Increase number of independent variable points.
            len_t += 1

        # Integration has stopped. Check if it was successful.
        if status == 1:
            success = True
        else:
            success = False

        if success:
            # Solution was successful.

            # Load values into output arrays.
            # The arrays built during integration likely have a bunch of unused junk at the end due to overbuilding their size.
            # This process will remove that junk and leave only the valid data.
            # These arrays will always be the same length or less (self.len_t <= new_size) than the ones they are
            # built off of, so it is safe to use Realloc.
            solution_t_ptr = <double *> reallocate_mem(
                time_domain_array_ptr,
                len_t * sizeof(double),
                'solution_t_ptr (success stage)')
            time_domain_array_ptr = NULL

            solution_y_ptr = <double_numeric *> reallocate_mem(
                y_results_array_ptr,
                y_size * len_t * sizeof(double_numeric),
                'solution_y_ptr (success stage)')
            y_results_array_ptr = NULL

            if capture_extra:
                solution_extra_ptr = <double_numeric *> reallocate_mem(
                    extra_array_ptr,
                    num_extra * len_t * sizeof(double_numeric),
                    'solution_extra_ptr (success stage)')
                extra_array_ptr = NULL
        else:
            # Clear the storage arrays used during the step loop
            if not (time_domain_array_ptr is NULL):
                free_mem(time_domain_array_ptr)
                time_domain_array_ptr = NULL
            if not (y_results_array_ptr is NULL):
                free_mem(y_results_array_ptr)
                y_results_array_ptr = NULL
            if capture_extra:
                if not (extra_array_ptr is NULL):
                    free_mem(extra_array_ptr)
                    extra_array_ptr = NULL

            # Integration was not successful. Leave the solution pointers as length 1 nan arrays.
            solution_t_ptr = <double *> allocate_mem(
                sizeof(double),
                'solution_t_ptr (fail stage)')
            solution_y_ptr = <double_numeric *> allocate_mem(
                y_size * sizeof(double_numeric),
                'solution_y_ptr (fail stage)')
            solution_extra_ptr = <double_numeric *> allocate_mem(
                num_extra * sizeof(double_numeric),
                'solution_extra_ptr (fail stage)')
            
            solution_t_ptr[0] = NAN
            for i in range(y_size):
                solution_y_ptr[i] = NAN
            for i in range(num_extra):
                solution_extra_ptr[i] = NAN

        # Integration is complete. Check if interpolation was requested.
        if success:
            if run_interpolation:
                # Use different len_t
                len_t_touse = len_t_eval
            else:
                len_t_touse = len_t
        else:
            # If integration was not successful use t_len = 1 to allow for nan arrays
            len_t_touse = 1

        if success and run_interpolation:
            # User only wants data at specific points.
            status = 2  # Interpolation is being performed.

            # TODO: The current version of cyrk_ode has not implemented sicpy's dense output. Instead we use an interpolation.
            # Build final interpolated time array
            interpolated_solution_t_ptr = <double *> allocate_mem(
                len_t_eval * sizeof(double),
                'interpolated_solution_t_ptr (interpolate stage)')

            # Build final interpolated solution arrays
            interpolated_solution_y_ptr = <double_numeric *> allocate_mem(
                y_size * len_t_eval * sizeof(double_numeric),
                'interpolated_solution_y_ptr (interpolate stage)')

            # Perform interpolation on y values
            interpolate(solution_t_ptr, t_eval_ptr, solution_y_ptr, interpolated_solution_y_ptr,
                        len_t, len_t_eval, y_size, y_is_complex)

            # Make a copy of t_eval (issues can arise if we store the t_eval pointer in solution array).
            for i in range(len_t_eval):
                interpolated_solution_t_ptr[i] = t_eval_ptr[i]

            if capture_extra:
                # Right now if there is any extra output then it is stored at each time step used in the RK loop.
                # We have to make a choice:
                #   - Do we interpolate the extra values that were stored?
                #   - Or do we use the interpolated t, y values to find new extra parameters at those specific points.
                # The latter method is more computationally expensive (recalls the diffeq for each y) but is more accurate.
                # This decision is set by the user with the `interpolate_extra` flag.

                # Build final interpolated solution array (Used if self.interpolate_extra is True or False)
                interpolated_solution_extra_ptr = <double_numeric *> allocate_mem(
                    num_extra * len_t_eval * sizeof(double_numeric),
                    'interpolated_solution_extra_ptr (interpolate)')

                if interpolate_extra:
                    # Perform interpolation on extra outputs
                    interpolate(solution_t_ptr, t_eval_ptr, solution_extra_ptr, interpolated_solution_extra_ptr,
                                len_t, len_t_eval, num_extra, y_is_complex)
                else:
                    # Use the new interpolated y and t values to recalculate the extra outputs with self.diffeq
                    for i in range(len_t_eval):
                        # Set state variables
                        t_now = t_eval_ptr[i]
                        for j in range(y_size):
                            y_view[j] = interpolated_solution_y_ptr[i * y_size + j]

                        # Call diffeq to recalculate extra outputs
                        if use_args:
                            diffeq(t_now, y_view, diffeq_out_view, *args)
                        else:
                            diffeq(t_now, y_view, diffeq_out_view)

                        # Capture extras
                        for j in range(num_extra):
                            interpolated_solution_extra_ptr[i * num_extra + j] = diffeq_out_view[extra_start + j]

                # Replace old pointers with new interpolated pointers and release the memory for the old stuff
                if not (solution_extra_ptr is NULL):
                    free_mem(solution_extra_ptr)
                solution_extra_ptr = interpolated_solution_extra_ptr
                interpolated_solution_extra_ptr = NULL

            # Replace old pointers with new interpolated pointers and release the memory for the old stuff
            if not (solution_t_ptr is NULL):
                free_mem(solution_t_ptr)
            solution_t_ptr = interpolated_solution_t_ptr
            interpolated_solution_t_ptr = NULL
            if not (solution_y_ptr is NULL):
                free_mem(solution_y_ptr)
            solution_y_ptr = interpolated_solution_y_ptr
            interpolated_solution_y_ptr = NULL

            # Interpolation is done.
            status = 1

        # Convert solution pointers to a more user-friendly numpy ndarray
        solution_t = np.empty(len_t_touse, dtype=np.float64, order='C')
        solution_y = np.empty((total_size, len_t_touse), dtype=DTYPE, order='C')
        solution_t_view = solution_t
        solution_y_view = solution_y

        for i in range(len_t_touse):
            solution_t_view[i] = solution_t_ptr[i]
            for j in range(y_size):
                solution_y_view[j, i] = solution_y_ptr[i * y_size + j]
        if capture_extra:
            for i in range(len_t_touse):
                for j in range(num_extra):
                    solution_y_view[extra_start + j, i] = solution_extra_ptr[i * num_extra + j]
        # free_mem solution arrays
        if not (solution_t_ptr is NULL):
            free_mem(solution_t_ptr)
            solution_t_ptr = NULL
        if not (solution_y_ptr is NULL):
            free_mem(solution_y_ptr)
            solution_y_ptr = NULL
        if capture_extra:
            if not (solution_extra_ptr is NULL):
                free_mem(solution_extra_ptr)
                solution_extra_ptr = NULL

        # Update integration message
        if status == 1:
            message = "Integration completed without issue."
        elif status == 0:
            message = "Integration is/was ongoing (perhaps it was interrupted?)."
        elif status == -1:
            message = "Error in step size calculation:\n\tRequired step size is less than spacing between numbers."
        elif status == -2:
            message = "Maximum number of steps (set by user) exceeded during integration."
        elif status == -3:
            message = "Maximum number of steps (set by system architecture) exceeded during integration."
        elif status == -6:
            message = "Integration never started: y-size is zero."
        elif status == -7:
            message = "Error in step size calculation:\n\tError in step size acceptance."

    finally:
        # free_mem pointers made from user inputs
        if not (tol_ptrs is NULL):
            free_mem(tol_ptrs)
            tol_ptrs = NULL
        if not (t_eval_ptr is NULL):
            free_mem(t_eval_ptr)
            t_eval_ptr = NULL

        # free_mem pointers used to track y, dydt, and any extra outputs
        if not (y_storage_ptrs is NULL):
            free_mem(y_storage_ptrs)
            y_storage_ptrs = NULL
        if not (extra_output_init_ptr is NULL):
            free_mem(extra_output_init_ptr)
            extra_output_init_ptr = NULL
        if not (extra_output_ptr is NULL):
            free_mem(extra_output_ptr)
            extra_output_ptr = NULL

        # free_mem RK Temp Storage Array
        if not (K_ptr is NULL):
            free_mem(K_ptr)
            K_ptr = NULL

        # free_mem other pointers that should have been freed in main loop, but in case of an exception they were missed.
        if not (solution_t_ptr is NULL):
            free_mem(solution_t_ptr)
            solution_t_ptr = NULL
        if not (solution_y_ptr is NULL):
            free_mem(solution_y_ptr)
            solution_y_ptr = NULL
        if not (solution_extra_ptr is NULL):
            free_mem(solution_extra_ptr)
            solution_extra_ptr = NULL
        if not (interpolated_solution_t_ptr is NULL):
            free_mem(interpolated_solution_t_ptr)
            interpolated_solution_t_ptr= NULL
        if not (interpolated_solution_y_ptr is NULL):
            free_mem(interpolated_solution_y_ptr)
            interpolated_solution_y_ptr = NULL
        if not (interpolated_solution_extra_ptr is NULL):
            free_mem(interpolated_solution_extra_ptr)
            interpolated_solution_extra_ptr = NULL
        if not (time_domain_array_ptr is NULL):
            free_mem(time_domain_array_ptr)
            time_domain_array_ptr = NULL
        if not (y_results_array_ptr is NULL):
            free_mem(y_results_array_ptr)
            y_results_array_ptr = NULL
        if not (extra_array_ptr is NULL):
            free_mem(extra_array_ptr)
            extra_array_ptr = NULL

    return solution_t, solution_y, success, message



from libc.string cimport memcpy
from libcpp cimport nullptr
from libcpp cimport bool as cpp_bool
from libcpp.cmath cimport fmin, fabs

cimport cpython.ref as cpy_ref
from CyRK.utils.vector cimport vector
from CyRK.utils.memory cimport shared_ptr, make_shared
from CyRK.cy.pysolver_cyhook cimport DiffeqMethod

cimport numpy as np

# =====================================================================================================================
# Import common functions and constants
# =====================================================================================================================
cdef extern from "common.cpp" nogil:
    const double INF
    const double EPS_100
    const unsigned int Y_LIMIT
    const unsigned int DY_LIMIT
    const double MAX_STEP

    ctypedef void (*DiffeqFuncType)(double*, double, double*, double*)

    cdef size_t find_expected_size(        
        int num_y,
        int num_extra,
        double t_delta_abs,
        double rtol_min)


cdef extern from "cy_array.cpp" nogil:
    size_t binary_search_with_guess(double key, const double* array, size_t length, size_t guess)


cdef extern from "dense.cpp" nogil:
    cdef cppclass CySolverDense:
        CySolverDense()
        CySolverDense(
            int integrator_int,
            double t_old,
            double t_now,
            double* y_in_ptr,
            unsigned int num_y,
            unsigned int Q_order)
        int integrator_int
        unsigned int num_y
        double t_old
        double t_now
        void call(double t_interp, double* y_interped)



# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef extern from "cysolution.cpp" nogil:
    cdef cppclass CySolverResult:
            CySolverResult()
            CySolverResult(
                const size_t num_y,
                const size_t num_extra,
                const size_t expected_size,
                const size_t last_t,
                const cpp_bool direction_flag,
                const cpp_bool capture_dense_output,
                const cpp_bool t_eval_provided)
            cpp_bool success
            cpp_bool reset_called
            char* message_ptr
            int error_code
            size_t size
            size_t num_y
            size_t num_dy
            cpp_bool capture_extra
            cpp_bool capture_dense_output
            cpp_bool t_eval_provided
            vector[double] time_domain
            vector[double] solution
            void save_data(double new_t, double* new_solution_y, double* new_solution_dy)
            void finalize()
            void reset()
            void update_message(const char* new_message)
            void call(const double t, double* y_interp)
            void call_vectorize(const double* t_array_ptr, size_t len_t, double* y_interp)

ctypedef shared_ptr[CySolverResult] CySolveOutput

cdef class WrapCySolverResult:
    """ Wrapper for the C++ class `CySolverResult` defined in "cysolution.cpp" """

    cdef shared_ptr[CySolverResult] cyresult_shptr
    cdef CySolverResult* cyresult_ptr
    cdef double* time_ptr
    cdef double* y_ptr
    cdef double[::1] time_view
    cdef double[::1] y_view

    cdef size_t size
    cdef size_t num_dy

    cdef void set_cyresult_pointer(self, shared_ptr[CySolverResult] cyresult_shptr)


# =====================================================================================================================
# Import CySolver Integrator Base Class
# =====================================================================================================================
cdef extern from "cysolver.cpp" nogil:
    cdef cppclass CySolverBase:
        CySolverBase()
        CySolverBase(
            DiffeqFuncType diffeq_ptr,
            shared_ptr[CySolverResult] storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval
        )
        
        shared_ptr[CySolverResult] storage_ptr
        int status
        int integration_method
        unsigned int num_y
        size_t len_t
        double t_now
        double* y0
        double* y_now_ptr
        double* dy_now_ptr
        cpp_bool check_status()
        void diffeq()
        void take_step()
        void change_storage(shared_ptr[CySolverResult] new_storage_ptr, cpp_bool auto_reset)
        void reset()
        void set_cython_extension_instance(cpy_ref.PyObject* cython_extension_class_instance)
        void py_diffeq()


cdef class WrapPyDiffeq:

    cdef object diffeq_func
    cdef tuple args
    cdef cpp_bool use_args
    cdef cpp_bool pass_dy_as_arg

    cdef unsigned int num_y
    cdef unsigned int num_dy

    cdef np.ndarray y_now_arr
    cdef double[::1] y_now_view
    cdef np.ndarray dy_now_arr
    cdef double[::1] dy_now_view

    # State attributes
    cdef double* y_now_ptr
    cdef double* t_now_ptr
    cdef double* dy_now_ptr

    cdef void set_state(self,
        double* dy_ptr,
        double* t_ptr,
        double* y_ptr
        ) noexcept
    
    cdef void diffeq(self) noexcept

# =====================================================================================================================
# Import CySolver Runge-Kutta Integrators
# =====================================================================================================================
cdef extern from "rk.cpp" nogil:
    cdef cppclass RKSolver(CySolverBase):
        RKSolver()
        RKSolver(
            DiffeqFuncType diffeq_ptr,
            shared_ptr[CySolverResult] storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void p_step_implementation()
        void p_estimate_error()
        void reset()
        void calc_first_step_size()


    cdef unsigned int RK23_METHOD_INT
    cdef cppclass RK23(RKSolver):
        RK23()
        RK23(
            DiffeqFuncType diffeq_ptr,
            shared_ptr[CySolverResult] storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void reset()
    
    cdef unsigned int RK45_METHOD_INT
    cdef cppclass RK45(RKSolver):
        RK45()
        RK45(
            DiffeqFuncType diffeq_ptr,
            shared_ptr[CySolverResult] storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void reset()
    
    cdef unsigned int DOP853_METHOD_INT
    cdef cppclass DOP853(RKSolver):
        DOP853()
        DOP853(
            DiffeqFuncType diffeq_ptr,
            shared_ptr[CySolverResult] storage_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool use_dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
        )
        void reset()


# =====================================================================================================================
# Import the C++ cysolve_ivp helper function
# =====================================================================================================================
cdef extern from "cysolve.cpp" nogil:
    # Pure C++ and Cython implementation
    cdef shared_ptr[CySolverResult] baseline_cysolve_ivp(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int method,
            const size_t expected_size,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size
            )

    # Python-hook implementation
    struct PySolverStatePointers:
        double* dy_now_ptr
        double* t_now_ptr
        double* y_now_ptr

    cdef cppclass PySolver:
        PySolver()
        PySolver(
            unsigned int integration_method,
            cpy_ref.PyObject* cython_extension_class_instance,
            DiffeqMethod cython_extension_class_diffeq_method,
            shared_ptr[CySolverResult] solution_ptr,
            const double t_start,
            const double t_end,
            const double* y0_ptr,
            const unsigned int num_y,
            const unsigned int num_extra,
            const double* args_ptr,
            const size_t max_num_steps,
            const size_t max_ram_MB,
            const cpp_bool dense_output,
            const double* t_eval,
            const size_t len_t_eval,
            const double rtol,
            const double atol,
            const double* rtols_ptr,
            const double* atols_ptr,
            const double max_step_size,
            const double first_step_size)
        PySolverStatePointers get_state_pointers()
        void solve()


# =====================================================================================================================
# Cython-based wrapper for baseline_cysolve_ivp that carries default values.
# =====================================================================================================================
cdef CySolveOutput cysolve_ivp(
    DiffeqFuncType diffeq_ptr,
    const double* t_span_ptr,
    const double* y0_ptr,
    const unsigned int num_y,
    unsigned int method = *,
    double rtol = *,
    double atol = *,
    double* args_ptr = *,
    unsigned int num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    double* t_eval = *,
    size_t len_t_eval = *,
    double* rtols_ptr = *,
    double* atols_ptr = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept nogil

cdef CySolveOutput cysolve_ivp_gil(
    DiffeqFuncType diffeq_ptr,
    const double* t_span_ptr,
    const double* y0_ptr,
    const unsigned int num_y,
    unsigned int method = *,
    double rtol = *,
    double atol = *,
    double* args_ptr = *,
    unsigned int num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    double* t_eval = *,
    size_t len_t_eval = *,
    double* rtols_ptr = *,
    double* atols_ptr = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

import numpy as np
np.import_array()

# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef class WrapCySolverResult:

    cdef void set_cyresult_pointer(self, shared_ptr[CySolverResult] cyresult_shptr):

        # Store c++ based result and pull out key information
        self.cyresult_shptr = cyresult_shptr
        self.cyresult_ptr   = cyresult_shptr.get()
        self.size           = self.cyresult_ptr[0].size
        self.num_dy         = self.cyresult_ptr[0].num_dy

        # Convert solution to pointers and views
        if self.cyresult_ptr.size > 0:
            self.time_ptr  = &self.cyresult_ptr.time_domain[0]
            self.y_ptr     = &self.cyresult_ptr.solution[0]
            self.time_view = <double[:self.size]>self.time_ptr
            self.y_view    = <double[:self.size * self.num_dy]>self.y_ptr
    
    def call(self, double t):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_ptr.capture_dense_output:
            raise AttributeError("Can not call WrapCySolverResult when dense_output set to False.")

        y_interp_array = np.empty(self.cyresult_ptr.num_y, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr      = &y_interp_view[0]

        self.cyresult_ptr.call(t, y_interp_ptr)
        return y_interp_array
    
    def call_vectorize(self, double[::1] t_view):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_ptr.capture_dense_output:
            raise AttributeError("Can not call WrapCySolverResult when dense_output set to False.")

        cdef size_t len_t = len(t_view)

        y_interp_array = np.empty(self.cyresult_ptr.num_y * len_t, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr = &y_interp_view[0]
        cdef double* t_array_ptr  = &t_view[0]

        self.cyresult_ptr.call_vectorize(t_array_ptr, len_t, y_interp_ptr)
        return y_interp_array.reshape(len_t, self.cyresult_ptr.num_y).T

    @property
    def success(self):
        return self.cyresult_ptr.success
        
    @property
    def message(self):
        return str(self.cyresult_ptr.message_ptr, 'UTF-8')
    
    @property
    def t(self):
        return np.asarray(self.time_view, dtype=np.float64, order='C')
    
    @property
    def y(self):
        return np.asarray(self.y_view, dtype=np.float64, order='C').reshape((self.size, self.num_dy)).T
    
    @property
    def size(self):
        return self.cyresult_ptr.size
    
    @property
    def error_code(self):
        return self.cyresult_ptr.error_code
    
    def __call__(self, t):

        if type(t) == np.ndarray:
            return self.call_vectorize(t)
        else:
            return self.call(t).reshape(self.cyresult_ptr.num_y, 1)

# =====================================================================================================================
# Create Wrapped cysolve_ivp (has various defaults)
# =====================================================================================================================

cdef CySolveOutput cysolve_ivp(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const unsigned int num_y,
            unsigned int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            double* args_ptr = NULL,
            unsigned int num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:
    
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

    return result

cdef CySolveOutput cysolve_ivp_gil(
            DiffeqFuncType diffeq_ptr,
            const double* t_span_ptr,
            const double* y0_ptr,
            const unsigned int num_y,
            unsigned int method = 1,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            double* args_ptr = NULL,
            unsigned int num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint dense_output = False,
            double* t_eval = NULL,
            size_t len_t_eval = 0,
            double* rtols_ptr = NULL,
            double* atols_ptr = NULL,
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept:
    
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_span_ptr,
        y0_ptr,
        num_y,
        method,
        expected_size,
        num_extra,
        args_ptr,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval,
        len_t_eval,
        rtol,
        atol,
        rtols_ptr,
        atols_ptr,
        max_step,
        first_step
        )

    return result

# =====================================================================================================================
# PySolver Class (holds the intergrator class and reference to the python diffeq function)
# =====================================================================================================================
cdef class WrapPyDiffeq:

    def __cinit__(
            self,
            object diffeq_func,
            tuple args,
            unsigned int num_y,
            unsigned int num_dy,
            bint pass_dy_as_arg = False
            ):
        
        # Install differential equation function and any additional args
        self.diffeq_func = diffeq_func
        if args is None:
            self.args     = None
            self.use_args = False
        else:
            if len(args) == 0:
                # Empty tuple provided. Don't use args.
                self.args     = None
                self.use_args = False
            else:
                self.args     = args
                self.use_args = True
        
        # Build python-safe arrays
        self.num_y  = num_y
        self.num_dy = num_dy

        if pass_dy_as_arg:
            self.pass_dy_as_arg = True
        else:
            self.pass_dy_as_arg = False
    
    cdef void set_state(self, double* dy_ptr, double* t_ptr, double* y_ptr) noexcept:
        self.dy_now_ptr = dy_ptr
        self.t_now_ptr  = t_ptr
        self.y_now_ptr  = y_ptr

        # Create memoryviews of the pointers
        self.y_now_view  = <double[:self.num_y]>self.y_now_ptr

        # Create numpy arrays which will be passed to the python diffeq.
        # We need to make sure that this is a not a new ndarray, but one that points to the same data. 
        # That is why we use `PyArray_SimpleNewFromData` instead of a more simple `asarray`.
        # Note that it is not safe to return these arrays outside of this class because they may get deallocated while
        # the numpy array still points to the underlying memory.
        cdef np.npy_intp[1] shape
        cdef np.npy_intp* shape_ptr = &shape[0]
        shape_ptr[0] = <np.npy_intp>self.num_y
        
        self.y_now_arr = np.PyArray_SimpleNewFromData(1, shape_ptr, np.NPY_DOUBLE, self.y_now_ptr)
        
        # Do the same for dy if the user provided the appropriate kind of differential equation.
        if self.pass_dy_as_arg:
            self.dy_now_view = <double[:self.num_dy]>self.dy_now_ptr
            shape[0]         = <np.npy_intp>self.num_dy  # dy may have a larger shape than y
            self.dy_now_arr  = np.PyArray_SimpleNewFromData(1, shape_ptr, np.NPY_DOUBLE, self.dy_now_ptr)   

    cdef void diffeq(self) noexcept:
        # Run python diffeq
        if self.pass_dy_as_arg:
            if self.use_args:
                self.diffeq_func(self.dy_now_arr, self.t_now_ptr[0], self.y_now_arr, *self.args)
            else:
                self.diffeq_func(self.dy_now_arr, self.t_now_ptr[0], self.y_now_arr)
        else:
            if self.use_args:
                self.dy_now_view = self.diffeq_func(self.t_now_ptr[0], self.y_now_arr, *self.args)
            else:
                self.dy_now_view = self.diffeq_func(self.t_now_ptr[0], self.y_now_arr)
            # Since we do not have a static dy that we can pass to the function and use in the solver we must copy over
            # the values from the newly created dy memory view
            # Note that num_dy may be larger than num_y if the user is capturing extra output during integration.
            memcpy(self.dy_now_ptr, &self.dy_now_view[0], sizeof(double) * self.num_dy)

# =====================================================================================================================
# PySolver wrapper function
# =====================================================================================================================

def pysolve_ivp(
        object py_diffeq,
        tuple time_span,
        const double[::1] y0,
        str method = 'RK45',
        const double[::1] t_eval = None,
        bint dense_output = False,
        tuple args = None,
        size_t expected_size = 0,
        unsigned int num_extra = 0,
        double first_step = 0.0,
        double max_step = INF,
        rtol = 1.0e-3,
        atol = 1.0e-6,
        size_t max_num_steps = 0,
        size_t max_ram_MB = 2000,
        bint pass_dy_as_arg = False
        ):

    # Parse method
    method = method.lower()
    cdef unsigned int integration_method = 1
    if method == "rk23":
        integration_method = 0
    elif method == "rk45":
        integration_method = 1
    elif method == 'dop853':
        integration_method = 2
    else:
        raise NotImplementedError(
            f"Unknown or unsupported integration method provided: {method}.\n"
            f"Supported methods are: RK23, RK45, DOP853."
            )
    
    # Parse time_span
    cdef double t_start = time_span[0]
    cdef double t_end   = time_span[1]
    cdef cpp_bool direction_flag = (t_end - t_start) >= 0

    # Parse y0
    cdef unsigned int num_y   = len(y0)
    cdef const double* y0_ptr = &y0[0]
    if num_y > Y_LIMIT:
        raise AttributeError(
            f"CyRK only supports a maximum number of {Y_LIMIT} dependent variables. {num_y} were provided."
            )
    
    # Parse t_eval
    cdef const double* t_eval_ptr = NULL
    cdef cpp_bool t_eval_provided = False
    cdef size_t len_t_eval = 0
    if t_eval is not None:
        t_eval_provided = True
        len_t_eval = len(t_eval)
        t_eval_ptr = &t_eval[0]
    
    # Parse num_extra
    if num_extra > (DY_LIMIT - Y_LIMIT):
        raise AttributeError(
            f"CyRK can only capture a maximum number of {DY_LIMIT - Y_LIMIT} extra outputs. {num_extra} were provided."
            )
    cdef unsigned int num_dy = num_y + num_extra
    
    # Parse rtol
    cdef double* rtols_ptr = NULL
    cdef double[::1] rtols_view
    cdef double rtol_float = 0.0
    if type(rtol) == float:
        rtol_float = rtol
        rtol_float = fabs(rtol_float)
    else:
        rtols_view = np.abs(np.asarray(rtol, dtype=np.float64, order='C'))
        rtols_ptr = &rtols_view[0]
    
    # Parse atol
    cdef double* atols_ptr = NULL
    cdef double[::1] atols_view
    cdef double atol_float = 0.0
    if type(atol) == float:
        atol_float = atol
        atol_float = fabs(atol_float)
    else:
        atols_view = np.abs(np.asarray(atol, dtype=np.float64, order='C'))
        atols_ptr = &atols_view[0]
    
    # Parse expected size
    cdef size_t expected_size_touse = expected_size
    cdef double rtol_tmp
    cdef double min_rtol = INF
    if expected_size_touse == 0:
        if rtols_ptr:
            # rtol for each y
            for y_i in range(num_y):
                rtol_tmp = rtols_ptr[y_i]
                if rtol_tmp < EPS_100:
                    rtol_tmp = EPS_100
                min_rtol = fmin(min_rtol, rtol_tmp)
        else:
            # Only one rtol
            rtol_tmp = rtol
            if rtol_tmp < EPS_100:
                rtol_tmp = EPS_100
            min_rtol = rtol_tmp
        expected_size_touse = find_expected_size(num_y, num_extra, fabs(t_end - t_start), min_rtol)
    
    # Parse first step size
    if first_step < 0.0:
        raise AttributeError("First step size must be a postive float (or 0.0 to use automatic finder).")

    # Parse maximum step size
    if max_step <= 0.0:
        raise AttributeError("Maximum step size must be a postive float.")

    # Build solution storage
    cdef shared_ptr[CySolverResult] result_ptr = make_shared[CySolverResult](
        num_y,
        num_extra,
        expected_size,
        t_end,
        direction_flag,
        dense_output,
        t_eval_provided)

    # Build diffeq wrapper
    cdef WrapPyDiffeq diffeq_wrap = WrapPyDiffeq(py_diffeq, args, num_y, num_dy, pass_dy_as_arg=pass_dy_as_arg)
    cdef DiffeqMethod diffeq_func = <DiffeqMethod>diffeq_wrap.diffeq

    # Finally we can actually run the integrator!
    # The following effectively copies the functionality of cysolve_ivp. We can not directly use that function
    # because we need to tie in the python-based diffeq function (via its wrapper)
    
    # Build null pointers to unused arguments
    cdef double* args_ptr = NULL

    # We need to heap allocate the PySolver class instance otherwise it can get garbage collected while the solver
    # is running.
    cdef PySolver* solver = new PySolver(
            integration_method,
            <cpy_ref.PyObject*>diffeq_wrap,
            diffeq_func,
            result_ptr,
            t_start,
            t_end,
            y0_ptr,
            num_y,
            num_extra,
            args_ptr,
            max_num_steps,
            max_ram_MB,
            dense_output,
            t_eval_ptr,
            len_t_eval,
            rtol_float,
            atol_float,
            rtols_ptr,
            atols_ptr,
            max_step,
            first_step
        )

    # Get pointers to the solver's state variables so that the Python differential equation can use and update them.
    cdef PySolverStatePointers state_pointers = solver.get_state_pointers()
    diffeq_wrap.set_state(
        state_pointers.dy_now_ptr,
        state_pointers.t_now_ptr,
        state_pointers.y_now_ptr
        )
    
    ##
    # Run the integrator!
    ##
    solver.solve()
    
    # Wrap the solution in a python-safe wrapper class
    cdef WrapCySolverResult pyresult = WrapCySolverResult()
    pyresult.set_cyresult_pointer(result_ptr)

    del solver

    return pyresult



import numpy as np
from numba import njit

from CyRK import cyrk_ode
from CyRK.cy.cysolvertest import CySolverExtraTest

@njit
def diffeq_extra_outputs(t, y, output):
    extra_0 = (1. - 0.01 * y[1])
    extra_1 = (0.02 * y[0] - 1.)
    output[0] = extra_0 * y[0]
    output[1] = extra_1 * y[1]
    output[2] = extra_0
    output[3] = extra_1


initial_conds = np.asarray((20., 20.), dtype=np.complex128)
initial_conds_float = np.asarray((20., 20.), dtype=np.float64)
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-7
atol = 1.0e-8


def test_extra_output_integration():
    """Check that the cython function solver is able to run and capture additional outputs """

    time_domain, all_output, success, message = \
        cyrk_ode(diffeq_extra_outputs, time_span, initial_conds, capture_extra=True, num_extra=2)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    assert all_output.dtype == np.complex128
    assert time_domain.size > 1
    assert time_domain.size == all_output[0].size
    assert len(all_output.shape) == 2
    assert all_output.shape[0] == 4
    assert all_output[0].size == all_output[1].size
    assert all_output[0].size == all_output[2].size
    assert all_output[0].size == all_output[3].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

def test_extra_output_integration_CySolver():
    """Check that the cython class solver is able to run and capture additional outputs """

    CySolverExtraTestInst = CySolverExtraTest(time_span, initial_conds_float, capture_extra=True, num_extra=2, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverExtraTestInst.t) == np.ndarray
    assert CySolverExtraTestInst.t.dtype == np.float64
    assert CySolverExtraTestInst.y.dtype == np.float64
    assert CySolverExtraTestInst.extra.dtype == np.float64
    assert CySolverExtraTestInst.t.size > 1
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.y[0].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.y[1].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.extra[0].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.extra[1].size
    assert CySolverExtraTestInst.y.shape[0] == 2
    assert CySolverExtraTestInst.extra.shape[0] == 2

    # Check that the other output makes sense
    assert type(CySolverExtraTestInst.success) == bool
    assert CySolverExtraTestInst.success
    assert type(CySolverExtraTestInst.message) == str

def test_extra_output_integration_teval_no_extra_interpolation():
    """Check that the cython function solver is able to run and capture additional outputs.
    Reduced t_eval used but no interpolation used on the extra parameters.
    """

    t_eval = np.linspace(time_span[0], time_span[1], 5)

    time_domain, all_output, success, message = \
        cyrk_ode(
            diffeq_extra_outputs, time_span, initial_conds, t_eval=t_eval,
            capture_extra=True, num_extra=2, interpolate_extra=False)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    assert all_output.dtype == np.complex128
    assert time_domain.size == t_eval.size
    assert time_domain.size == all_output[0].size
    assert len(all_output.shape) == 2
    assert all_output.shape[0] == 4
    assert all_output[0].size == all_output[1].size
    assert all_output[0].size == all_output[2].size
    assert all_output[0].size == all_output[3].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

def test_extra_output_integration_teval_no_extra_interpolation_CySolver():
    """Check that the cython class solver is able to run and capture additional outputs.
    Reduced t_eval used but no interpolation used on the extra parameters.
    """

    t_eval = np.linspace(time_span[0], time_span[1], 5)

    CySolverExtraTestInst = CySolverExtraTest(time_span, initial_conds_float, t_eval=t_eval,
                                              capture_extra=True, num_extra=2, interpolate_extra=False, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverExtraTestInst.t) == np.ndarray
    assert CySolverExtraTestInst.t.dtype == np.float64
    assert CySolverExtraTestInst.y.dtype == np.float64
    assert CySolverExtraTestInst.extra.dtype == np.float64
    assert CySolverExtraTestInst.t.size > 1
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.y[0].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.y[1].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.extra[0].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.extra[1].size
    assert CySolverExtraTestInst.y.shape[0] == 2
    assert CySolverExtraTestInst.extra.shape[0] == 2

    # Check that the other output makes sense
    assert type(CySolverExtraTestInst.success) == bool
    assert CySolverExtraTestInst.success
    assert type(CySolverExtraTestInst.message) == str

def test_extra_output_integration_teval_with_extra_interpolation():
    """Check that the cython function solver is able to run and capture additional outputs
    Reduced t_eval used with interpolation used on the extra parameters.
    """

    t_eval = np.linspace(time_span[0], time_span[1], 5)

    time_domain, all_output, success, message = \
        cyrk_ode(
                diffeq_extra_outputs, time_span, initial_conds, t_eval=t_eval,
                capture_extra=True, num_extra=2, interpolate_extra=True)

    # Check that the ndarrays make sense
    assert type(time_domain) == np.ndarray
    assert time_domain.dtype == np.float64
    assert all_output.dtype == np.complex128
    assert time_domain.size == t_eval.size
    assert time_domain.size == all_output[0].size
    assert len(all_output.shape) == 2
    assert all_output.shape[0] == 4
    assert all_output[0].size == all_output[1].size
    assert all_output[0].size == all_output[2].size
    assert all_output[0].size == all_output[3].size

    # Check that the other output makes sense
    assert type(success) == bool
    assert success
    assert type(message) == str

def test_extra_output_integration_teval_with_extra_interpolation_CySolver():
    """Check that the cython class solver is able to run and capture additional outputs
    Reduced t_eval used with interpolation used on the extra parameters.
    """

    t_eval = np.linspace(time_span[0], time_span[1], 5)

    CySolverExtraTestInst = CySolverExtraTest(time_span, initial_conds_float, t_eval=t_eval,
                                              capture_extra=True, num_extra=2, interpolate_extra=True, auto_solve=True)

    # Check that the ndarrays make sense
    assert type(CySolverExtraTestInst.t) == np.ndarray
    assert CySolverExtraTestInst.t.dtype == np.float64
    assert CySolverExtraTestInst.y.dtype == np.float64
    assert CySolverExtraTestInst.extra.dtype == np.float64
    assert CySolverExtraTestInst.t.size > 1
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.y[0].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.y[1].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.extra[0].size
    assert CySolverExtraTestInst.t.size == CySolverExtraTestInst.extra[1].size
    assert CySolverExtraTestInst.y.shape[0] == 2
    assert CySolverExtraTestInst.extra.shape[0] == 2

    # Check that the other output makes sense
    assert type(CySolverExtraTestInst.success) == bool
    assert CySolverExtraTestInst.success
    assert type(CySolverExtraTestInst.message) == str



# distutils: language = c
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
import cython

from libc.math cimport fmax, fmin, floor
from libc.math cimport INFINITY as INF
from libc.float cimport DBL_EPSILON as EPS
from libc.stdint cimport SIZE_MAX, INT32_MAX
from libc.stdlib cimport free

from CyRK.utils.utils cimport allocate_mem, reallocate_mem, free_mem
from CyRK.array.interp cimport interp_array_ptr, interp_complex_array_ptr

# # Integration Constants
# Multiply steps computed from asymptotic behaviour of errors by this.
cdef double SAFETY           = 0.9
cdef double MIN_FACTOR       = 0.2  # Minimum allowed decrease in a step size.
cdef double MAX_FACTOR       = 10.  # Maximum allowed increase in a step size.
cdef double MAX_STEP         = INF
cdef double EPS_10           = EPS * 10.
cdef double EPS_100          = EPS * 100.
cdef size_t MAX_SIZET_SIZE = <size_t>(SIZE_MAX)
cdef size_t MAX_INT_SIZE   = <size_t>(INT32_MAX)

# # Memory management constants
# Assume that a cpu has a L1 of 300KB. Say that this progam will have access to 75% of that total.
cdef double CPU_CACHE_SIZE = 0.75 * 300_000.
# Number of entities we can fit into that size is based on the size of double (or double complex)
cdef double MAX_ARRAY_PREALLOCATE_SIZE_DBL      = 600_000.
cdef double MAX_ARRAY_PREALLOCATE_SIZE_DBLCMPLX = 300_000.
cdef double MIN_ARRAY_PREALLOCATE_SIZE = 10.
cdef double ARRAY_PREALLOC_TABS_SCALE  = 1000.  # A delta_t_abs higher than this value will start to grow array size.
cdef double ARRAY_PREALLOC_RTOL_SCALE  = 1.0e-5  # A rtol lower than this value will start to grow array size.
# RAM_BUFFER_SIZE should be set to the max size we expect cyrk_ode or CySolver to be before integration starts.
#  i.e., before the solution arrays start to grow.
# As of CyRK v0.8.3 CySolver is around 1200 bytes. Buffer this up to 2000.
# Note this does not need to be precise. It just should be close.
cdef size_t RAM_BUFFER_SIZE = 2000


cdef void interpolate(
        double* time_domain_full,
        double* time_domain_reduced,
        double_numeric* target_array_full,
        double_numeric* target_array_reduced,
        size_t t_len_full,
        size_t t_len_reduced,
        size_t target_len,
        bint is_complex
        ) noexcept nogil:
    """ Interpolate the results of a successful integration over the user provided time domain, `time_domain_full`. """

    # Setup loop variables
    cdef size_t i, j

    # Build a pointer array that will contain only 1 y for all ts in time_domain_full
    cdef double_numeric* array_slice_ptr = <double_numeric *> allocate_mem(
        t_len_full * sizeof(double_numeric),
        'array_slice_ptr (common.interpolate)')

    # Build a pointer that will store the interpolated values for 1 y at a time; size of self.len_t_eval
    cdef double_numeric* interpolated_array_slice_ptr = <double_numeric *> allocate_mem(
        t_len_reduced * sizeof(double_numeric),
        'interpolated_array_slice_ptr (common.interpolate)')
    
    try:
        for j in range(target_len):
            # The interpolation function only works on 1D arrays, so we must loop through each of the y variables.
            # # Set timeslice equal to the time values at this y_j
            for i in range(t_len_full):
                # OPT: Inefficient memory looping
                array_slice_ptr[i] = target_array_full[i * target_len + j]

            # Perform numerical interpolation
            if double_numeric is cython.doublecomplex:
                interp_complex_array_ptr(
                    time_domain_reduced,
                    time_domain_full,
                    array_slice_ptr,
                    interpolated_array_slice_ptr,
                    t_len_full,
                    t_len_reduced)
            else:
                interp_array_ptr(
                    time_domain_reduced,
                    time_domain_full,
                    array_slice_ptr,
                    interpolated_array_slice_ptr,
                    t_len_full,
                    t_len_reduced)

            # Store result.
            for i in range(t_len_reduced):
                # OPT: Inefficient memory looping
                target_array_reduced[i * target_len + j] = interpolated_array_slice_ptr[i]
    finally:
        # Release memory of any temporary variables
        if not (array_slice_ptr is NULL):
            free(array_slice_ptr)
            array_slice_ptr = NULL
        if not (interpolated_array_slice_ptr is NULL):
            free(interpolated_array_slice_ptr)
            interpolated_array_slice_ptr = NULL

cdef size_t find_expected_size(
        size_t y_size,
        size_t num_extra,
        double t_delta_abs,
        double rtol_min,
        bint capture_extra,
        bint is_complex) noexcept nogil:

    cdef double temp_expected_size
    # Pick starting value that works with most problems
    temp_expected_size = 500.0
    # If t_delta_abs is very large or rtol is very small, then we may need more. 
    temp_expected_size = \
        fmax(
            temp_expected_size,
            fmax(
                t_delta_abs / ARRAY_PREALLOC_TABS_SCALE,
                ARRAY_PREALLOC_RTOL_SCALE / rtol_min
                )
            )
    # Fix values that are very small/large
    temp_expected_size = fmax(temp_expected_size, MIN_ARRAY_PREALLOCATE_SIZE)

    if is_complex:
        max_expected = MAX_ARRAY_PREALLOCATE_SIZE_DBL
    else:
        max_expected = MAX_ARRAY_PREALLOCATE_SIZE_DBLCMPLX
    if capture_extra:
        max_expected /= (y_size + num_extra)
    else:
        max_expected /= y_size
    
    temp_expected_size = fmin(temp_expected_size, max_expected)
    # Store result as int
    cdef size_t expected_size_to_use = <size_t>floor(temp_expected_size)
    return expected_size_to_use


cdef void find_max_num_steps(
        size_t y_size,
        size_t num_extra,
        size_t max_num_steps,
        size_t max_ram_MB,
        bint capture_extra,
        bint is_complex,
        bint* user_provided_max_num_steps,
        size_t* max_num_steps_touse) noexcept nogil:

    # Determine max number of steps
    cdef double max_num_steps_ram_dbl
    max_num_steps_ram_dbl = max_ram_MB * (1000 * 1000)
    # As of CyRK v0.8.3, the CySolver class takes up about 1200 Bytes of memory. Let's assume cyrk_ode takes up a 
    #  similar amount.
    # Buffer the expeceted size up a bit (set by RAM_BUFFER_SIZE) and subtract this from the total we are allowed.
    max_num_steps_ram_dbl -= <double> RAM_BUFFER_SIZE
    # Divide by size of data that will be stored in main loop
    if is_complex:
        max_num_steps_ram_dbl /= sizeof(double complex)
    else:
        max_num_steps_ram_dbl /= sizeof(double)
    # Divide by number of dependnet and extra variables that will be stored. The extra "1" is for the time domain.
    if capture_extra:
        max_num_steps_ram_dbl /= (1 + y_size + num_extra)
    else:
        max_num_steps_ram_dbl /= (1 + y_size)
    cdef size_t max_num_steps_ram = <size_t> floor(max_num_steps_ram_dbl)

    # Parse user-provided max number of steps
    user_provided_max_num_steps[0] = False
    if max_num_steps == 0:
        # No user input; use ram-based value
        max_num_steps_touse[0] = max_num_steps_ram
    else: 
        if max_num_steps > max_num_steps_ram:
            max_num_steps_touse[0] = max_num_steps_ram
        else:
            user_provided_max_num_steps[0] = True
            max_num_steps_touse[0] = max_num_steps
    # Make sure that max number of steps does not exceed size_t limit
    if max_num_steps_touse[0] > (MAX_SIZET_SIZE / 10):
        max_num_steps_touse[0] = (MAX_SIZET_SIZE / 10)